'****************************************************************
'**   PROGRAM CURVEFIT  (QUICKBASIC 4.5)                       **
'**   Written 10/23/90 by Milton Watts                         **
'**   Revised 1/25/91  to support extended coefficient files   **
'**   Revised 10/21/92 to include temperature coefficient file **
'**   Revised 4/13/93  to include .CRT and .CRF extionsions    **
'**                                                            **
'**   Computes 2 dimensional polynomial LMS curvefit           **
'**   coefficients for QUARTZDYNE pressure transducer          **
'**   calibration.  Output files have the following extensions **
'**         .OUT         Fit Error File                        **
'**         .CFF/.CRF    Primary (PRESSURE) Coefficients       **
'**         .CFT/.CRT    Secondary (TEMPERATURE) Coefficients  **
'****************************************************************

' The following is a compiler directive allowing the use of REDIM
' $DYNAMIC

DEFDBL A-Z
DEFINT I-N
OPTION BASE 1

DECLARE FUNCTION dosfile$ (file$, ext$)
DECLARE FUNCTION numstr$ (a#)
DECLARE SUB Prescale (x#(), m#, X0#, npts%)
DECLARE SUB Polyfit (x#(), y#(), z#(), c#(), r#(), npts%, s1%, s2%)
DECLARE SUB Solve (a#(), b#(), c#(), s%, errflag%)
DECLARE SUB minmax (xmin#, xmax#, x#(), npts%)

fext$ = "CFF"     ' default coefficient file format
maxlen = 200     ' longest allowed filename

file$ = RTRIM$(LTRIM$(UCASE$(COMMAND$)))
i = INSTR(file$, "/R")         ' look for Reference Based Extension Option
IF i > 0 THEN                  ' /R option is present
  fext$ = "CRF"
  IF i = 1 THEN                ' /R is before filename
    file$ = MID$(file$, 2)       ' remove it
  ELSE
    file$ = LEFT$(file$, i - 1)   ' remove /R from after filename
  END IF
END IF
file$ = RTRIM$(LTRIM$(file$))
IF file$ = "" THEN  ' no filename given
  PRINT "USAGE: CURVEFIT datafile[.DAT] [/R]"
  PRINT
  END
END IF
IF INSTR(file$, ".") = 0 THEN  ' if no extension specified assume .DAT
  file$ = file$ + ".DAT"
END IF

'  read data file  -  assume 4 columns seperated by commas
'  first row contains model and calibration date
'  next three rows contain name, crystal id, and units respectively
'  Any NULL data must have a comma placeholders

OPEN file$ FOR INPUT AS #1

REDIM pres(maxlen), temp(maxlen), fpres(maxlen), ftemp(maxlen), dp(maxlen), dt(maxlen)

INPUT #1, model$, caldate$
INPUT #1, plabel$, tlabel$, fplabel$, ftlabel$
INPUT #1, pid$, tid$, fpid$, ftid$
INPUT #1, punit$, tunit$, fpunit$, ftunit$

IF INSTR(fext$, "R") = 1 THEN   ' this is a reference fit
  IF INSTR(fpid$, "R") = 0 THEN ' Make sure not already added
    fpid$ = fpid$ + "R"         ' indicate with "R" in 1st line of CFFILE 
  END IF
END IF


i = 0
WHILE NOT EOF(1)
  i = i + 1
  INPUT #1, pres(i), temp(i), fpres(i), ftemp(i)
WEND
CLOSE #1
npts = i

PRINT npts; " DATA POINTS READ"
PRINT plabel$; " FIT:"
INPUT "ENTER FIT ORDER IN PRESSURE FREQUENCY (3) > ", n1p
INPUT "ENTER FIT ORDER IN TEMPERATURE FREQUENCY (3) > ", n2p
np = (n1p + 1) * (n2p + 1)
REDIM pc(np)

PRINT tlabel$; " FIT:"
INPUT "ENTER FIT ORDER IN PRESSURE FREQUENCY (0) > ", n1t
INPUT "ENTER FIT ORDER IN TEMPERATURE FREQUENCY (3) > ", n2t
nt = (n1t + 1) * (n2t + 1)
REDIM tc(nt)

scale = .01#
fp0 = INT(fpres(1))
ft0 = INT(ftemp(1))

CALL Prescale(fpres(), scale, fp0, npts)
CALL Prescale(ftemp(), scale, ft0, npts)
CALL minmax(tmin, tmax, temp(), npts)
CALL minmax(pmin, pmax, pres(), npts)

IF np > 1 THEN
CALL Polyfit(fpres(), ftemp(), pres(), pc(), dp(), npts, n1p, n2p)
file$ = dosfile$(file$, fext$)
PRINT plabel$; " COEFFICIENT FILE: "; file$;
PRINT
PRINT "FIT ORDER IN "; ftlabel$; ":"; n2p
PRINT "Prescale Algorithm:  Xt = Mt * (Ft - Ft0)"
PRINT "Ft0 ="; ft0
PRINT "Mt  ="; .01
PRINT
PRINT "FIT ORDER IN "; fplabel$; ":"; n1p
PRINT "Prescale Algorithm:  Xp = Mp * (Fp - Fp0)"
PRINT "Fp0 ="; fp0
PRINT "Mp  ="; .01
PRINT
FOR i = 1 TO np
  PRINT "  C("; i; ") = ", numstr$(pc(i))
NEXT i
PRINT

'  store pressure coefficient file in file.CFF or file.CRF

OPEN file$ FOR OUTPUT AS #1
PRINT #1, fpid$         ' crystal id
PRINT #1, plabel$      ' output label (PRESSURE)
PRINT #1, punit$        ' calibration units (psi)
PRINT #1, n2p            ' fit order in temperature frequency
PRINT #1, 1             ' prescale algorithm: 1 = m * (f - f0)
PRINT #1, numstr$(scale)' value of m for ft
PRINT #1, numstr$(ft0)  ' value of f0 for ft
PRINT #1, n1p            ' fit order in pressure frequency
PRINT #1, 1             ' prescale algorithm for fp
PRINT #1, numstr$(scale)' value of m for fp
PRINT #1, numstr$(fp0)  ' value of f0 form fp
FOR i = 1 TO np
  PRINT #1, numstr$(pc(i))   ' calibration coefficients
NEXT i
PRINT #1, 1              ' span
PRINT #1, 0              ' zero
PRINT #1, numstr$(tmin)  ' minimum temperature
PRINT #1, numstr$(tmax)  ' maximum temperature
PRINT #1, numstr$(pmin)  ' minimum pressure
PRINT #1, numstr$(pmax)  ' maximum pressure
PRINT #1, caldate$
PRINT #1, model$
CLOSE #1
END IF

IF (nt > 1) THEN
CALL Polyfit(fpres(), ftemp(), temp(), tc(), dt(), npts, n1t, n2t)
MID$(fext$, 3) = "T"
file$ = dosfile$(file$, fext$)
PRINT tlabel$; " COEFFICIENT FILE: "; file$;
PRINT
PRINT "FIT ORDER IN "; ftlabel$; ":"; n2t
PRINT "Prescale Algorithm:  Xt = Mt * (Ft - Ft0)"
PRINT "Ft0 ="; ft0
PRINT "Mt  ="; .01
PRINT
PRINT "FIT ORDER IN "; fplabel$; ":"; n1t
PRINT "Prescale Algorithm:  Xp = Mp * (Fp - Fp0)"
PRINT "Fp0 ="; fp0
PRINT "Mp  ="; .01
PRINT
FOR i = 1 TO nt
  PRINT "  C("; i; ") = ", numstr$(tc(i))
NEXT i
PRINT

'  store temperature coefficient file in file.CFT or file.CRT

OPEN file$ FOR OUTPUT AS #1
PRINT #1, fpid$         ' crystal id
PRINT #1, tlabel$      ' output label (TEMPERATURE)
PRINT #1, tunit$        ' calibration units ('C)
PRINT #1, n2t            ' fit order in temperature frequency
PRINT #1, 1             ' prescale algorithm: 1 = m * (f - f0)
PRINT #1, numstr$(scale)' value of m for ft
PRINT #1, numstr$(ft0)  ' value of f0 for ft
PRINT #1, n1t            ' fit order in pressure frequency
PRINT #1, 1             ' prescale algorithm for fp
PRINT #1, numstr$(scale)' value of m for fp
PRINT #1, numstr$(fp0)  ' value of f0 form fp
FOR i = 1 TO nt
  PRINT #1, numstr$(tc(i))   ' calibration coefficients
NEXT i
PRINT #1, 1              ' span
PRINT #1, 0              ' zero
PRINT #1, numstr$(tmin)  ' minimum temperature
PRINT #1, numstr$(tmax)  ' maximum temperature
PRINT #1, numstr$(pmin)  ' minimum pressure
PRINT #1, numstr$(pmax)  ' maximum pressure
PRINT #1, caldate$
PRINT #1, model$
CLOSE #1
END IF

'  store residuals in file.out
IF np > 1 OR nt > 1 THEN
file$ = dosfile$(file$, "OUT")
OPEN file$ FOR OUTPUT AS #1
WRITE #1, plabel$, tlabel$, LEFT$(plabel$, 1) + " ERROR", LEFT$(tlabel$, 1) + " ERROR"
WRITE #1, pid$, tid$, fpid$, ftid$
WRITE #1, punit$, tunit$, punit$, tunit$
FOR i = 1 TO npts
  PRINT #1, numstr$(pres(i)); ","; numstr$(temp(i)); ","; numstr$(dp(i)); ","; numstr$(dt(i))
NEXT i
CLOSE #1
END IF
END

REM $STATIC
DEFSNG A-H, O-Z
FUNCTION dosfile$ (file$, ext$)
  i = INSTR(file$, ".")
  IF i > 0 THEN
    dosfile$ = LEFT$(file$, i) + ext$
  ELSE
    dosfile$ = file$ + "." + ext$
  END IF
END FUNCTION

DEFDBL A-H, O-Z
SUB minmax (xmin, xmax, x(), npts)
  xmin = x(1)
  xmax = x(1)
  FOR i = 2 TO npts
    IF x(i) < xmin THEN xmin = x(i)
    IF x(i) > xmax THEN xmax = x(i)
  NEXT i
END SUB

FUNCTION numstr$ (a)
  a$ = STR$(a)           ' convert number to string
  i = INSTR(a$, "D")     ' look for D notation for exponents
  IF i > 0 THEN
    MID$(a$, i) = "E"    ' change D to E for compatability
  END IF
  numstr$ = a$
END FUNCTION

'
SUB Polyfit (x(), y(), z(), c(), r(), npts, n1, n2)
  DIM errflag AS INTEGER
  n = (n1 + 1) * (n2 + 1)
  REDIM f(npts, n), a(n), b(n, n), binv(n, n)
  '
  '   CREATE POWERS OF XY MATRIX F(npts,n)
  '
  FOR k = 1 TO npts
    f(k, 1) = 1#
    FOR i = 1 TO n2
      f(k, i + 1) = f(k, i) * y(k)
    NEXT i
    FOR i = i TO n - 1
      f(k, i + 1) = f(k, i - n2) * x(k)
    NEXT i
  NEXT k
  '
  '  Create A[n] = Z[npts]*F[npts,n]
  '
  FOR i = 1 TO n
    a(i) = 0#
    FOR k = 1 TO npts
      a(i) = a(i) + z(k) * f(k, i)
    NEXT k
  NEXT i
  '
  ' CREATE B[n,n] = TRN( F[npts,n] ) F[npts,n]
  '
  FOR i = 1 TO n
    FOR j = 1 TO i
      b(i, j) = 0#
      FOR k = 1 TO npts
        b(i, j) = b(i, j) + f(k, i) * f(k, j)
      NEXT k
      b(j, i) = b(i, j)
    NEXT j
  NEXT i
  '
  '  Solve matrix equation A[n] = B[n,n]C[n] for C[n]
  '
  CALL Solve(a(), b(), c(), n, errflag)
  IF errflag <> 0 THEN ' MATRIX SOLUTION DIDN'T WORK
    BEEP
    PRINT "Ill conditioned data, no solution was found'"
    STOP
  END IF
  '
  '  COMPUTE FIT RESIDUALS
  '
  FOR k = 1 TO npts
    r(k) = 0#
    FOR i = 1 TO n
      r(k) = r(k) + f(k, i) * c(i)
    NEXT i
    r(k) = r(k) - z(k)
  NEXT k
END SUB

DEFDBL I-N
SUB Prescale (x(), m, X0, npts%)
  DIM n AS INTEGER
  FOR n = 1 TO npts%
    x(n) = m * (x(n) - X0)
  NEXT n
END SUB

DEFINT I-N
'
SUB Solve (a(), b(), c(), n, errflag AS INTEGER)
  REDIM d(n)

  '
  '  DECOMPOSE B (LDLt GAUSSIAN ELIMINATION)
  '  Assumes B is a symmetric matrix - A and B are modified
  '
  errflag = 0
  IF b(1, 1) = 0 THEN GOTO Errexit
  d(1) = 1 / b(1, 1)
  FOR i = 2 TO n
    b(i, 1) = b(i, 1) * d(1)
    NEXT i
    FOR i = 2 TO n
      FOR j = 1 TO i - 1
        b(i, i) = b(i, i) - b(j, j) * b(i, j) * b(i, j)
      NEXT j
      IF b(i, i) = 0 THEN GOTO Errexit
      d(i) = 1 / b(i, i)
      FOR j = i + 1 TO n
        FOR k = 1 TO i - 1
          b(j, i) = b(j, i) - b(k, k) * b(i, k) * b(j, k)
        NEXT k
        b(j, i) = b(j, i) * d(i)
      NEXT j
    NEXT i
    '
    '  REDUCE COLUMN VECTOR
    '
    FOR i = 2 TO n
      FOR j = 1 TO i - 1
        a(i) = a(i) - b(i, j) * a(j)
      NEXT j
    NEXT i
    '
    '  BACKSUBSTITE TO OBTAIN C
    '
    FOR i = 1 TO n
      c(i) = a(i) * d(i)
    NEXT i
    FOR i = n - 1 TO 1 STEP -1
      FOR j = i + 1 TO n
        c(i) = c(i) - c(j) * b(j, i)
      NEXT j
    NEXT i
    EXIT SUB
Errexit:    'Zero Diagonal Term found - Can't solve with Gauss elimination
    errflag = 1
END SUB

