/**********************************************************
**    CURVEFIT.C  Written 3/13/91 by MHW                 **
**    Computes Coefficients for Quartzdyne transducers   **
***********************************************************/

#include <stdlib.h>
#include <stdio.h>
//#include <io.h>
#include <conio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "curvefit.h"

int main (int argc, char *argv[] ){
	cfrec cf;
	char infile[128], outfile[128];
	FILE *fptr;
	double *xp=NULL,*xt=NULL;

	enum ColIndx {INDX=0, PRES, TEMP, Fp, Ft, Fr, R1, R2, R3, R4};
	double *cd[10]={ // array of pointers to arrays for column data
			NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};

	char row1[1024];// first row of header data
	colLabel cl[10];// name, units, id for each column (rows 2-4 of header)

	int npts, i, nCols=6;
	int Npp, Npt, Ntp, Ntt;
	int isRef;

	const char *help =
"\n"
"Usage: CURVEFIT datafile outfile Npp Npt [Ntp Ntt] [/Mscale] [/Ooffset]\n"
"  datafile - Source data filepath (.DAT extension assumed if not specified)\n"
"      This is a Quote and Comma deliminated file as provided by Quartzdyne\n"
"  outfile - Output filepath (.CFF .CFT .CRF .CRT and .OUT generated)\n"
"  Npp - Calcualted Pressure fit order in pressure frequency\n"
"  Npt - Calcualted Pressure fit order in temperature frequency\n"
"  Ntp - Calcualted Temperature fit order in pressure frequency\n"
"  Ntt - Calcualted Temperature fit order in temperature frequency\n"
"  /Mscale - Alternate prescaling factor (0.01 assumed)\n"
"  /Ooffset - Alternate prescaling offset (First datapoint assumed)\n\n"
"Example:\n"
"  >CURVEFIT A:\\123456.DAT A:\\123456_1 3 3 0 3 /M1.0 /O0.0\n"
"\n";

	// set defaults for prescale
	cf.p1 = 1;
	cf.p2 = 1;
	cf.f01 = -9998;// flag to indicate not specified
	cf.f02 = -9998;
	cf.m1 = 0.01;
	cf.m2 = 0.01;
	cf.span = 1.0;
	cf.zero = 0.0;

	Ntp = 0;
	Ntt = 0;

	// process command line arguments
	npts = 0; // counts required arguments
	for (i=1;i<argc;) {
		if (*argv[i] == '/') { // process optional arguments in any order
			switch (argv[i][1]) {
			case 'M': case 'm': // scale
				cf.m2 = cf.m1 = atof(argv[i++]+2);
				if (cf.m2 == 0.0) {
					printf("Invalid value for \M switch <%s>/n/n",argv[i]);
					printf(help);
					return(1);
				}
				break;
			case 'O': case 'o': // offset
					cf.f01 = cf.f02 = atof(argv[i++]+2);
				break;
			default:
				printf("Unknown switch <%s>\n\n",argv[i]);
				printf(help);
				return(1);
			}
		} else switch(++npts) {  // process required arguments in order
		case 1:
			strcpy(infile,argv[i++]);
			break;
		case 2:
			strcpy(outfile,argv[i++]);
			break;
		case 3:
			Npp = atoi(argv[i++]);
			break;
		case 4:
			Npt = atoi(argv[i++]);
			break;
		case 5:
			Ntp = atoi(argv[i++]);
			break;
		case 6:
			Ntt = atoi(argv[i++]);
			break;
		case 7:
			printf(help);
			return(1);
		}
	}
	if (npts < 4) { // must specify at least infile, outfile, Npp, Ntp
		printf(help);
		return(1);
	}

/* open data file */

	if ( (fptr = fopen (infile, "r" ) ) == NULL ) {
		dosextension(infile,"DAT");
		if ( (fptr = fopen (infile, "r" ) ) == NULL ) {
			fprintf(stderr,"\nError opening file %s\n",infile);
			return(1);
		}
	}
	if (read_header(fptr, row1, cl, nCols) != 0 ) {
		printf("Error reading data file header rows\n");
		return(1);
	}
	npts = count_lines(fptr );
	rewind(fptr);
	if (npts <= 0) {
		printf("No data found in file");
		exit(1);
	}
	i= max( (Npp+1)*(Npt+1), (Ntp+1)*(Ntt+1) );
	if (dalloc(&cf.c,i) == NULL) {
		printf("Not enough memory\n");
		exit(1);
	}
	if ( dalloc(&xp,npts) == NULL	|| dalloc(&xt,npts) == NULL) {
		printf("Not enough memory\n");
		free(cf.c);
		if (xt != NULL) free(xp);
		if (xt != NULL) free(xt);
		exit(1);
	}
	for (i=0;i<10;++i) {
		if ( dalloc( &(cd[i]), npts ) == NULL ) {
			printf("Not enough memory\n");
			// cleanup
			free(cf.c);
			free(xp);
			free(xt);
			while (--i>0)	free(cd[i]);
			exit(1);
		}
	}
	if (read_header(fptr, row1, cl, nCols) != 0 ) {
		printf("Incorrect file format reading header\n");
		// cleanup
		free(cf.c);
		free(xp);
		free(xt);
		for (i=0;i<10;++i) free(cd[i]);
		exit(1);
	};
	if ( read_data(fptr, cd, nCols) != npts) {
		printf("Incorrect file format reading data\n");
		// cleanup
		free(cf.c);
		free(xp);
		free(xt);
		for (i=0;i<10;++i) free(cd[i]);
		exit(1);
	};
	getFileDate(cf.date, fptr);// get cal date from DOS filestamp
	fclose(fptr);

	// set up cf struct for calculations
	stripQuotes(cf.id, cl[Ft].id);       // get serial number
	stripQuotes(cf.model, cl[Fp].id);    // get model number
	getRange(&cf.tmin, &cf.tmax, cd[TEMP], 5, npts);
	getRange(&cf.pmin, &cf.pmax, cd[PRES], 1000, npts);
	if (cf.f01 == -9998) { // no F0 specified, take first point from file
		cf.f01 = floor(*cd[Fp]+.5);
		cf.f02 = floor(*cd[Ft]+.5);
	}

	// repeat for different coefficient types

	do_prescale(xp, cd[Fp], cf.p1, cf.m1, cf.f01, npts);
	do_prescale(xt, cd[Ft], cf.p2, cf.m2, cf.f02, npts);

	// look for 'R' in serial number to indicate reference-based frequencies
	isRef = ( strchr(cf.id,'R')!=NULL || strchr(cf.id,'r') != NULL );

	if (!isRef) {  // compute standard coefficients
		if (Npp + Npt >0) {
			cf.n1 = Npp;
			cf.n2 = Npt;
			stripQuotes(cf.name, cl[PRES].name);// Pressure
			stripQuotes(cf.units, cl[PRES].units);// psi
			sprintf( cl[nCols].name, "\"DIFF %s\"", cf.name);
			sprintf( cl[nCols].units, "\"%s\"", cf.units);
			sprintf( cl[nCols].id, "\"%s\"", "STANDARD");
			if ( (i = polyfit(xp,xt,cd[PRES],cd[nCols],cf.c,npts,cf.n1,cf.n2) ) != 0) {
				printf("Error %d using polyfit\nUnable to compute fit",i);
			} else if ( (fptr = outfopen(outfile,"CFF") ) != NULL ) {
				store_cf(fptr,&cf);
				close(fptr);
				printf("Created %s\n",outfile);
			} else {
				printf ("Unable to store coefficients in %s\n",outfile);
			}
			nCols++;
		}
		if (Ntp + Ntt > 0) {
			cf.n1 = Ntp;
			cf.n2 = Ntt;
			stripQuotes(cf.name, cl[TEMP].name);// Temperature
			stripQuotes(cf.units, cl[TEMP].units);// C
			sprintf( cl[nCols].name, "\"DIFF %s\"", cf.name);
			sprintf( cl[nCols].units, "\"%s\"", cf.units);
			sprintf( cl[nCols].id, "\"%s\"", "STANDARD");

			if ( (i = polyfit(xp,xt,cd[TEMP],cd[nCols],cf.c,npts,cf.n1,cf.n2) ) != 0) {
				printf("Error %d using polyfit\nUnable to compute fit",i);
			} else if ( (fptr = outfopen(outfile,"CFT") ) != NULL ) {
				store_cf(fptr,&cf);
				close(fptr);
				printf("Created %s\n",outfile);
			} else {
				printf ("Unable to store coefficients in %s\n",outfile);
			}
			nCols++;
		}
		// recompute prescale for reference-based coefficients
		if (do_ref_prescale(xp, cd[Fp], cd[Fr], cf.p1, cf.m1, cf.f01, npts) == 0) {
			do_ref_prescale(xt, cd[Ft], cd[Fr], cf.p2, cf.m2, cf.f02, npts);
			strcat(cf.id,"R");
			isRef = !isRef; // if succesful, we are set up to do reference fit
		}
	}
	// Now calculate reference based coefficients

	if (isRef) {
		if (Npp + Npt >0) {
			cf.n1 = Npp;
			cf.n2 = Npt;
			stripQuotes(cf.name, cl[PRES].name);// Pressure
			stripQuotes(cf.units, cl[PRES].units);// psi
			sprintf( cl[nCols].name, "\"DIFF %s\"", cf.name);
			sprintf( cl[nCols].units, "\"%s\"", cf.units);
			sprintf( cl[nCols].id, "\"%s\"", "REFRENCE-BASED");

			if ( (i = polyfit(xp,xt,cd[PRES],cd[nCols],cf.c,npts,cf.n1,cf.n2) ) != 0) {
				printf("Error %d using polyfit\nUnable to compute fit",i);
			} else if ( (fptr = outfopen(outfile,"CRF") ) != NULL ) {
				store_cf(fptr,&cf);
				close(fptr);
				printf("Created %s\n",outfile);
			} else {
				printf ("Unable to store coefficients in %s\n",outfile);
			}
			nCols++;
		}

		if (Ntp + Ntt >0) {
			cf.n1 = Ntp;
			cf.n2 = Ntt;
			stripQuotes(cf.name, cl[TEMP].name);// Temperature
			stripQuotes(cf.units, cl[TEMP].units);// C
			sprintf( cl[nCols].name, "\"DIFF %s\"", cf.name);
			sprintf( cl[nCols].units, "\"%s\"", cf.units);
			sprintf( cl[nCols].id, "\"%s\"", "REFERENCE-BASED");

			if ( (i = polyfit(xp,xt,cd[TEMP],cd[nCols],cf.c,npts,cf.n1,cf.n2) ) != 0) {
				printf("Error %d using polyfit\nUnable to compute fit\n",i);
			} else if ( (fptr = outfopen(outfile,"CRT") ) != NULL ) {
				store_cf(fptr,&cf);
				close(fptr);
				printf("Created %s\n",outfile);
			} else {
				printf ("Unable to store coefficients in %s\n",outfile);
			}
			nCols++;
		}
	}

// done computing fits, now store residuals

	if ( (fptr = outfopen(outfile,"OUT") ) != NULL ) {
		store_header(fptr, row1, cl, nCols);
		store_data(fptr, cd, nCols, npts);
		close(fptr);
		printf("Created %s\n",outfile);
	} else {
		printf ("Unable to store data in %s\n",outfile);
	}

	// cleanup
	free(cf.c);
	free(xp);
	free(xt);
	for (i=0;i<10;++i) free(cd[i]);
	return(0);
}

void store_cf(FILE *fp, cfrec *cf) {
		int ncf;
		double *cptr;

		fprintf (fp, "%s\n%s\n%s\n", cf->id, cf->name, cf->units);
		fprintf (fp, "% d\n% d\n", cf->n2,cf->p2);
		fprintf (fp, "% .12G\n% .12G\n", cf->m2, cf->f02);
		fprintf (fp, "% d\n% d\n", cf->n1,cf->p1);
		fprintf (fp, "% .12G\n% .12G\n", cf->m1, cf->f01);
		ncf = (cf->n1+1)*(cf->n2+1);       /* Total number of coefficients */
		cptr= cf->c;
		while (ncf--) fprintf (fp, "% .12G\n",*cptr++);

		fprintf(fp,"% .12G\n% .12G\n",cf->span,cf->zero);
		fprintf(fp,"% .12G\n% .12G\n",cf->tmin,cf->tmax);
		fprintf(fp,"% .12G\n% .12G\n",cf->pmin,cf->pmax);
		fprintf(fp,"%s\n%s\n",cf->date,cf->model);
}

char *fgetl(char *s, int n, FILE *fp) {
		/* gets a line from fp trimming leading and traling white space
			 and any characters exceeding n in count */

		int c;
		char *cs,*d;
		cs=d=s;
		while ( (c = getc(fp)) != EOF && ( isspace(c) ));  /* eat leading white */
		if (c == EOF) {
				*s = '\0';
				return (NULL);
		}
		ungetc(c,fp);
		while ( (c = getc(fp)) != EOF && c != '\n')
				if ( --n>0 && ( ! isspace(*cs++ = c) ) ) d = cs;
		*d = '\0';
		return( s );
}

int count_lines (FILE * fp){
		char line[1024];
		int n=0;
		while ( fgetl(line, 1024, fp) != NULL)
			if (*line != '*') ++n;  // use '*' to eliminate points
		return(n);
}

int read_header(FILE *fp, char *row1, colLabel *cl, int nCols ) {
	char line[2048];
	char *str;
	int i,j;

	if ( fgetl(row1, 2048, fp) == NULL) return(1);

	if ( fgetl(line, 2048, fp) == NULL) return(2);
	if ( (str = strtok(line, ",\t")) == NULL ) return(2);
	strcpy(cl[0].name, str);
	for (j=1;j<nCols;++j) {
		if ( (str = strtok(NULL, ",\t")) == NULL ) return(2);
		strcpy(cl[j].name, str);
	}

	if ( fgetl(line, 2048, fp) == NULL) return(3);
	if ( (str = strtok(line, ",\t")) == NULL ) return(3);
	strcpy(cl[0].units, str);
	for (j=1;j<nCols;++j) {
		if ( (str = strtok(NULL, ",\t")) == NULL ) return(3);
		strcpy(cl[j].units, str);
	}

	if ( fgetl(line, 2048, fp) == NULL) return(4);
	if ( (str = strtok(line, ",\t")) == NULL ) return(4);
	strcpy(cl[0].id, str);
	for (j=1;j<nCols;++j) {
		if ( (str = strtok(NULL, ",\t")) == NULL ) return(4);
		strcpy(cl[j].id, str);
	}

	return(0);
}

void getRange(double *rmin, double *rmax, double *dptr, double rsln, int npts) {
	*rmin = *rmax = *dptr++;
	while (--npts > 0) {
		if (*dptr < *rmin) *rmin = *dptr;
		else if (*dptr > *rmax) *rmax = *dptr;
		dptr++;
	}
	if (rsln != 0.0) {
		*rmin = rsln * floor (*rmin/rsln + 0.5);
		*rmax = rsln * floor (*rmax/rsln + 0.5);
	}
}

int getFileDate(char *dest, FILE *fp) {
	const char *monthText[]={
			 "JAN","FEB","MAR","APR","MAY","JUN",
			 "JUL","AUG","SEP","OCT","NOV","DEC"
	};
	struct ftime {
		unsigned ft_tsec  : 5;  /* Two seconds */
		unsigned ft_min   : 6;  /* Minutes */
		unsigned ft_hour  : 5;  /* Hours */
		unsigned ft_day   : 5;  /* Days */
		unsigned ft_month : 4;  /* Months */
		unsigned ft_year  : 7;  /* Year - 1980 */
	} ft;

	if ( getftime(fileno(fp), &ft) == 0 ) {    // get date from data file
		sprintf(dest,"%u %s %u", ft.ft_day, monthText[ft.ft_month-1],ft.ft_year+1980);
		return(0);
	};
	return (-1);
}

int store_header(FILE *fp, char *row1, colLabel *cl, int nCols ) {
	int i;

	fprintf(fp,"%s\n%s",row1,cl[0].name);
	for (i=1;i<nCols;++i) fprintf(fp,",%s",cl[i].name);
	fprintf(fp,"\n%s",cl[0].units);
	for (i=1;i<nCols;++i) fprintf(fp,",%s",cl[i].units);
	fprintf(fp,"\n%s",cl[0].id);
	for (i=1;i<nCols;++i) fprintf(fp,",%s",cl[i].id);
	fprintf(fp,"\n");

	return(0);
}

int read_data(FILE *fp, double *cd[], int nCols) {
	char line[2048],*str;
	int n=0,i=0, errFlag;
	while (fgetl(line , 2048, fp) != NULL) {
		if (*line != '*') {   // '*' indicates eliminate point
			errFlag = 0;
			if ( *(str = strtok(line, " ,\t\"")) != NULL ) {
				cd[0][n] = atof(str);
				for (i = 1;i<nCols;++i) {
					if ( *(str = strtok(NULL, " ,\t\"")) != NULL ) {
						cd[i][n] = atof(str);
					} else {
						errFlag = -1;
						continue;
					}
				}
			}
			if (errFlag == 0) ++n;// increment only if entire line OK.
		}
	}
	return(n);
}

int store_data(FILE *fp, double *cd[], int nCols, int nPts) {
	int i,j;

	for (i=0;i<nPts;++i) {
		fprintf(fp,"%3.0f",cd[0][i]);
		for (j=1;j<nCols;++j) fprintf(fp,",%.12G",cd[j][i]);
		fprintf(fp,"\n");
	}
	return(0);
}

void stripQuotes(char *dest, char *src) {
	while (*src != '\0') {
		if (*src != '"') *dest++ = *src++;
		else src++;
	}
	*dest= '\0';
}

double * dalloc ( double **x, int npts) {
		if ( *x != NULL) free(*x);
		*x = (double *) calloc( npts, sizeof(double) );
		return(*x);
}

FILE * outfopen(char *file, char *ext) {
		dosextension (file, ext);
		return (fopen(file,"w"));
}

void dosextension (char *file, char *ext) {
		char c;
		while ( (c = *file) != '.' && c != '\0') file++;
		*file++ = '.';
		if (*ext == '.') ext++;
		while( (*file++ = *ext++) != '\0');
		return;
}

void do_prescale(
				double *xp, double *x, int type, double m, double f0, int npts) {
		while (npts-- >0) *xp++ = prescale( *x++, type, m, f0);
}

int do_ref_prescale(double *xp, double *x, double *ref,
				int type, double m, double f0, int npts) {
		while (npts-- >0) {
			if (*ref == 0.0) return (-1);
			*xp++ = prescale( 7.2e6 * *x++ / *ref++, type, m, f0);
		}
		return (0);
}


double prescale( double f, int type, double m, double f0) {
		switch(type) {
		case 1:
				return(m * ( f - f0 ) );
		case 2:
				if (f != 0.0) return(m*(1-f0/f));
				return (0.0);
		case 3:
				if (f != 0.0) return (m*(1-( f0*f0/(f*f) ) ) );
				return (0.0);
		default:
				return(f);
		}
}

int polyfit(double *x, double *y, double *z, double *r,
						double *c, int n, int s1, int s2) {
		int s, flag;
		double *a, **b, *d, **f;

		if (s1<0 || s2<0) return (1); /* arguments out of range */
		s = (s1+1)*(s2+1);

    if (alloctemps (&a,&b,&d,&f,n,s) != 0) return(2);/* memory full */
		dopowers(f, x, y, n, s1, s2, s); /* create polynomial terms for each n */
    matmult_12(a, z, f, n, s);       /*  lh vect a[s] = z[n] * f[n][s]     */
    crossterms(b, f, n, s);          /*  b[s][s] = ftrn[s][n] * f[n][s]    */
		if ( (flag = solve(a, b, c, d, s)) == 0 ) {
        matmult_21(r, f, c, n, s);   /*  output r[n] = f[n][s] * c[s]      */
        matsub( r, r, z, n);         /*  resid r[n] = r[n] - z[n]          */
    }
		freetemps(a,b,d,f,n,s);
    return(flag);
}

int alloctemps(double **a, double **b[], double **d, double **f[],
        int n, int s){
    int i,errflag = 0;
    if( (*a = (double * ) calloc(s,sizeof(double  )) ) == NULL ) errflag = 1;
    if( (*b = (double **) calloc(s,sizeof(double *)) ) == NULL ) errflag = 1;
		if( (*d = (double * ) calloc(s,sizeof(double  )) ) == NULL ) errflag = 1;
    if( (*f = (double **) calloc(n,sizeof(double *)) ) == NULL ) errflag = 1;
    for (i=s;i>0;--i)
        if( ( (*b)[i-1] = (double *) calloc (i,sizeof(double)) )
								== NULL ) errflag = 1;
    for (i=0;i<n;++i)
        if( ( (*f)[i] = (double *) calloc (s,sizeof(double)) )
                == NULL ) errflag = 1;
    if(errflag) freetemps(*a,*b,*d,*f,n,s);
		return(errflag);
}

void freetemps(double *a, double *b[], double *d, double *f[],
        int n,int s){
    while (--n >= 0) free(f[n]);
    free(f);
    free(d);
    while (--s >= 0) free(b[s]);
    free(b);
    free(a);
}

void dopowers(double *f[], double *x, double *y,
				int n, int s1, int s2, int s) {

    /***********************************************************************
    *   Compute powers of x and y, placing them in f where:                *
    *   f[n][0,...,s-1] =                                                  *
    *       = {x[n]^0 * y[n]^0 , x[n]^0 * y[n]^1 , ... , x[n]^s1 * y[n]^s2}*
    *   f is an 'n' lengthed array of pointers to rows of 's' doubles      *
    ***********************************************************************/

    register double *f1ptr,*f2ptr;
    while (n--) {
        *(f2ptr = *f) = 1.0;
        f1ptr = f2ptr+1;
        for ( s1=1; s1 <= s2; ++s1) *f1ptr++ = *f2ptr++ * *y;
				f2ptr = *f++;
        for (  ; s1<s; ++s1) *f1ptr++ = *f2ptr++ * *x;
        x++;
				y++;
    }
}

void matmult_12(double *a, double * b, double *c[], int npts, int s) {
    /* a[s] = b[n]*c[n][s]  */
    register int n,i;
    double * bn;
    for (i=0; i<s; ++i) {
        bn = b;
        *a = 0.0;
				for (n=0; n<npts; ++n) *a += *bn++ * c[n][i];
        a++;
    }
}

void matsub( double *a, double *b, double *c, int n) {
    while (n--) *a++ = *b++ - *c++;
}

void matmult_21 ( double * r, double *f[], double *c, int npts, int s) {
       /* r[n] = f[n][s] * c[s] */
    register int n,i;
    register double *fni, *ci;
		for (n=0; n<npts; ++n) {
				*r = 0.0;
        fni = *f++;
        ci = c;
        for (i=0; i<s; ++i) *r += *fni++ * *ci++;
        r++;
    }
}

void crossterms( double *bi[], double *f[], int npts, int s) {

    /*********************************************************************
    **  b[s][s] = trn(f[n][s]) * f[n][s]                                **
    **  since b[i][j] == b[j][i] for all valid i and j, only the        **
    **  lower triangle matrix will be computed and stored               **
		*********************************************************************/
		register int i,j,n;
    register double *bij;
    for (i=0; i<s; ++i, ++bi) {
        bij = *bi;
				for (j=0; j<=i; ++j, ++bij) {
            *bij = 0.0;
            n=npts;
            while (n--) *bij += f[n][i] * f[n][j];
        }
    }
}

int solve(double *a, double *b[], double *c, double *d, int s) {
		register int i, j, k;

    /*  decompose B - LDLt Gauss Elimination  */
    double *bi, *bj, *bii, *di;

    bi = *b;
    if (*bi == 0.0) return (3);/* invalid matrix */
		di = d;
    *di = 1.0/ *bi;
    for (i=1;i<s;++i) *b[i] *= *di;
    for (i=1;i<s;++i) {
				bi = b[i];
        bii = bi+i;
        for (j=0; j<i; j++) *bii -= b[j][j] * bi[j] * bi[j];
				if (*bii == 0.0) return (3); /* invalide matrix */
        *(++di) = 1.0/ *bii;
        for ( j=i+1; j<s; ++j) {
            bj = b[j];
            bii = bj+i;
            for (k=0; k<i; ++k) {
                *bii -= b[k][k] * bi[k] * bj[k];
            }
            *bii *= *di;
        }
    }

    /*  reduce column vector  */

		for (i=1;i<s;++i) {
        bi = b[i];
        di = a+i;
				for (j=0;j<i;++j) {
            *di -= bi[j] * a[j];
        }
    }

    /* Back substitute to obtain C */

    for (i=0; i<s; ++i) c[i] = a[i]*d[i];
    for (i=s-2; i>=0; i--) {
        bi = c+i;
        for (j=i+1;j<s;++j) {
						*bi -= c[j] * b[j][i];
        }
    }
    return(0);
}
