/**********************************************************************
**   ROMREAD.C                                                       **
**   Example program using Interupt Driven Serial Port Routines      **
**   Written 4/28/97 by Milton Watts                                 **
**   QUARZDYNE, Inc.                                                 **
**                                                                   **
**   Reads commands from a file such as created by ROMREAD           **
**   and sends them to a Series I                                    **
**                                                                   **
**                                                                   **
**********************************************************************/


#include <time.h>   /* clock(), CLK_TIC */
#include <bios.h>   /* _bios_equiplist */
#include <string.h> /* strtok() */
#include <stdio.h>  /* printf() etc */

#include "serial.h"

int getlineNS(FILE *fp, char *line) ;
int getcmd(FILE *fp, char *cmd) ;
int getver(char *ptr) ;
int wait_for_message(int seconds);

#define MAXCHAR 1024

int main(int argc, char * argv[]) {
	FILE *fp;
	int port, nports, echo=0;
	char *ptr, reply[MAXCHAR],cmd[MAXCHAR] ;
	char address[]="01";
	int i,nfiles=0,c;
	char cmdfile[40]="SIDAT.TXT";
	char usage[]=	"\nUsage: ROMWRITE COMn [#nn] [/e] [/bnnnn] [cmdfile]"
						"     COMn indicates the serial port (n=1 or 2)\n"
						"     [#nn] is the logical device interface address (default is 01)\n"
						"     [/e] allows for RS485 interfaces which echo outgoing characters\n"
						"     [/bnnnn] allows for baud rate setting nnnn=300|600|1200|2400|4800|9600\n"
						"     [cmdfile] sets the command file path (default is SIDAT.TXT)\n"
						"     brackets indicate optional parameters\n"
	;
	unsigned baud = BAUD_1200;  // default baud rate

	if (argc < 2 || strnicmp (argv[1],"COM",3) != 0) {
		printf(usage);
		return 1;
	}
	port = argv[1][3]-'1';
	nports = ( _bios_equiplist() & 0x0E00 ) >> 9;
	if (port < 0 || port >=nports ) {
		printf("\nCommunications port %s not found\n",argv[1]);
		return 1;
	}

	for (i=2;i<argc;++i) {        /* read optional parameters */
		if ( *argv[i] == '#' && strlen(argv[i]) == 3 )  { // address specifier
			strcpy(address, argv[i]+1);
		} else if ( *argv[i] == '/' )  { // slash option
			switch ( *(argv[i]+1) ) {
			case 'E':
			case 'e':
				echo = 1;
				break;
			case 'B':
			case 'b':
				switch(*(argv[i]+2) ) {
				case '1':
					baud=_COM_1200;
					break;
				case '2':
					baud=_COM_2400;
					break;
				case '3':
					baud=_COM_300;
					break;
				case '4':
					baud=_COM_4800;
					break;
				case '6':
					baud=_COM_600;
					break;
				case '9':
					baud=_COM_9600;
					break;
				default:
					printf("\nInvalid baud rate specifier %s\n",argv[i]);
					printf(usage);
					exit(1);
				}
				break;
			}
		} else if (nfiles<1) {  // parameter must be input file specifier
			strcpy(cmdfile,argv[i]);
			nfiles++;
		} else {
			printf(usage);
			return 1;
		};
	};

	/* first make sure the communications link is open */
	baud |= COMM_N81;  // settings = baud + No Parity + 8 Data bits + 1 stop	baud |= COMM_N81;
	if ( comm_init(port, baud, echo)==0 ) {
		printf("\nCommunications Port Cannot be initialized\n");
		printf("%s\n",error_message( comm_error(0) ) );
		return 1;
	}

	send_message(address,"Ver");  /* check communications-get ROM version */
	if ( wait_for_message(2) == 0 ) {
		printf("\n\n\nNo response on COM%d address #%s\n", port+1, address);
		printf("Check cables, device address and COMn setting\n");
		comm_restore();
		return 1;
	}
	rcv_message(reply);
	printf(reply);
	ptr=strstr(reply,"Ver ");
	if (ptr == NULL) {
		printf("\nInvalid response on COM%d address #%s\n", port+1, address);
		printf("Check cables, device address and COMn setting\n");
		comm_restore();
		return 1;
	}

	if ( (fp=fopen(cmdfile,"r")) == NULL ) {
		printf("Unable to open command data file <%s>\n",cmdfile);
		comm_restore();
		return 1;
	}

	while (getcmd(fp,cmd) ) {    // read command from command file
		send_message(address,cmd);  // query instrument
		if ( wait_for_message(10) == 0 ) {
			printf("\nNo response on COM%d address #%s to command\n", port+1, address);
			printf("%s\n",cmd);
			comm_restore();
			fclose(fp);
			return 1;
		}
		rcv_message(reply);
		printf(reply);
		if (strchr(reply,('{'))!= NULL) { // multiline command
			while (wait_for_message(1) !=0 && strchr(reply,('}')) == NULL) {
				rcv_message(reply);		 // receive remaining lines of reply
				printf(reply);
			}
		}
	}

	comm_restore();
	fclose(fp);
	return(0);
}

int getlineNS(FILE *fp, char *line) {
// gets one line of text from fp, trimming whitespace and comments
	int c, nchars=0;
	while ( ( c=getc(fp) ) != EOF) {
		switch(c) {
		case ';': case '!':
			while (  (c=getc(fp) ) != EOF && c != '\n');//plow thru comment
			if (nchars>0) {  // if we've got something, return it
				line[nchars++]='\n';
				line[nchars]='\0';
				return nchars;
			}  // else check on next line
			break;
		case ' ':case'\t':
			break; // ignore white space
		case '\n':
			if (nchars>0) {  // if we've got something, return it
				line[nchars++]='\n';
				line[nchars]='\0';
				return nchars;
			}  // else check on next line
			break;
		default:
			line[nchars++]=c;
		}
	}
	if (nchars>0) line[nchars++]='\n';
	line[nchars]='\0';// return what we have
	return nchars;
}

int getcmd(FILE *fp, char *cmd) {
	int c,nchars=0;

	if ( (c=getlineNS(fp,cmd) ) > 0 ) {
		if (strchr(cmd,'{')!= NULL) { // multiline command
			nchars+=c;
			while ( (c=getlineNS(fp,cmd+nchars) ) > 0   // append next line
						&& strchr(cmd+nchars,'}') == NULL) {          // look for '}'
				nchars+=c;                                // count total characters
			} // next line
		}
		return nchars+c;
	} // EOF
	return (0);
}

int getver(char *ptr) {
	int ver;  // convert "n.nn" to integer between 1 and 999
	ver = 100 * (ptr[0]-'0')+ 10 * (ptr[2]-'0') +ptr[3]-'0';
	return ver;
}

int wait_for_message(int seconds) {
	/* monitors message_rcvd() until either a message
		has been detected or until the number of seconds has elapsed.
		returns 1 if a message is detected, 0 otherwise.
	*/
	clock_t timeout;

	timeout= clock() + ( clock_t ) ( seconds * CLK_TCK ) ;
		 /* timeout in clock tics since program start */
	while ( message_rcvd(0) <=0 ) {  /* check for message first */
		if ( clock() > timeout ) return 0;
	}
	return 1;
}

